﻿// ***********************************************************************************
// CameraBackgroundColor
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// Licensed under CC0 - https://creativecommons.org/publicdomain/zero/1.0/
//
// Version 5:
// * Added experimental quick toggle for renderers.
//   Idea based on post by hijku:
//   https://hub.virtamate.com/threads/vd-for-passthrough.43272/post-135631
//
// Version 4:
// * Merged session- and scene-plugin versions.
// * A warning will now show if multiple CameraBackgroundColor script where loaded (excludes check for old Session Plugin)
// * Fixed error for scenes without WindowCamera atom
// * Fixed OnEnable()-NullReferenceException appearing in VAMs logfile
// * Fixed logic with the UI related to enabled / disabled state. The UI is no longer accessible in disabled state.
//   In disabled state it's no longer possible to set a color or toggle the skybox with scene triggers or the UI.

using System;
using SimpleJSON;
using UnityEngine;
using MVR.FileManagementSecure;
using System.Collections.Generic;
using System.Linq;
using System.Collections;

public class CameraBackgroundColor : MVRScript
{
    private bool debugToggleDefaultState = false;
    private Camera mainCam;
    private Camera screenshotCam;
    private Camera thumbnailCam;
    private Camera windowCam;
    private JSONStorableColor jColor;
    private JSONStorableBool jShowSkybox;
    private JSONStorableBool jOriginalSkyboxState;
    private JSONStorableBool jRenderers;
    private JSONStorableBool jDebug;
    private SkyshopLightControllerUI skyUI;
    private const string SAVE_FOLDER = "Saves\\PluginData\\CameraBackgroundColor";
    private const string SAVE_FILENAME = "default";
    private const string SAVE_EXTENSION = "json";
    private bool initOK = false;
    private bool isStarted = false;
    private bool isEnabled = false;
    private UIDynamicColorPicker colorPicker;
    private UIDynamicButton buttonSave;
    private UIDynamicButton buttonLoad;
    private UIDynamicButton buttonDelete;
    private UIDynamicToggle toggleRenderers;
    private UIDynamicToggle toggleSkybox;
    private UIDynamicTextField textField;
    private const string TEXT = "To prevent the skybox from blocking the view on the <i>actual</i> background it is automatically disabled by default.\n\nTo auto-load this with VAM:\n1) Load it as [Session Plugin]\n2) go to [Session Plugin Presets]\n > [Change User Defaults...]\n > [Set Current As User Defaults]\n\nVersion 6 (CC0)";
    private string[] alwaysVisibleAtomTypes = { "Person", "WindowCamera", "CoreControl", "VRController", "Dildo", "Padle", "ToyAH", "ToyBP", "Button", "UIButton", "UIButtonImage", "UISlider", "UIToggle" };
    private class Filter {
        public string atom;
        public JSONStorableBool show;
        public UIDynamicToggle toggle;
    }
    private List<Filter> filters;

    public override void Init()
    {
        try
        {
            CheckForMultipleLoadedScripts();
            mainCam = CameraTarget.centerTarget.targetCamera;
            screenshotCam = SuperController.singleton.hiResScreenshotCamera;
            thumbnailCam = SuperController.singleton.screenshotCamera;
            Atom windowCamAtom = GetSceneAtoms().Find((Atom a) => { return a.type == "WindowCamera"; });
            if (windowCamAtom != null)
            {
                CameraControl cameraControl = windowCamAtom.GetStorableByID("CameraControl") as CameraControl;
                if (cameraControl != null)
                    windowCam = cameraControl.cameraToControl;
            }

            // UI LEFT
            // Colorpicker
            jColor = new JSONStorableColor("Color", HSVColorPicker.RGBToHSV(0f, 0f, 1f), SetColor);
            RegisterColor(jColor);
            colorPicker = CreateColorPicker(jColor);
            // Save Button
            buttonSave = CreateButton("Save Session Plugin Color");
            buttonSave.button.onClick.AddListener(SavePluginDataJSON);
            // Load Button
            buttonLoad = CreateButton("Load Session Plugin Color");
            buttonLoad.button.onClick.AddListener(LoadPluginDataJSON);
            // Delete Button
            buttonDelete = CreateButton("Delete Session Plugin Color");
            buttonDelete.button.onClick.AddListener(DeletePluginDataJSON);
            // Skybox Toggle - get on/off skybox state from VAM UI & store original state, from SkyshopLightController.InitUI()
            skyUI = SkyshopLightController.singleton.UITransform.GetComponentInChildren<SkyshopLightControllerUI>();
            jOriginalSkyboxState = new JSONStorableBool("OriginalSkyboxState", skyUI.showSkyboxToggle.isOn);
            RegisterBool(jOriginalSkyboxState);
            jShowSkybox = new JSONStorableBool("Show Skybox", false);
            RegisterBool(jShowSkybox);
            toggleSkybox = CreateToggle(jShowSkybox);

            jShowSkybox.setCallbackFunction = ShowSkybox;
            skyUI.showSkyboxToggle.isOn = jShowSkybox.val; // toggle skybox on/off
            // Renderers Toggle
            jRenderers = new JSONStorableBool("Render [Control + R] <color=#FF0000>(experimental)</color>", true);
            RegisterBool(jRenderers);
            toggleRenderers = CreateToggle(jRenderers);
            jRenderers.setCallbackFunction = ShowRenderers;
            // Debug Toggle
            jDebug = new JSONStorableBool("Debug messages", debugToggleDefaultState);
            CreateToggle(jDebug).height = 10f;
            // Patreon button
            UIDynamicButton button = CreateButton("patreon.com/sallyw");
            button.button.onClick.AddListener(OnClickSupportButton);
            // Text
            textField = CreateTextField(new JSONStorableString("Text", TEXT));
            textField.height = 352f;

            initOK = true;
        }
        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
    }

    void Start()
    {
        try
        {
            isStarted = true;
            if (containingAtom.type == "SessionPluginManager")
                LoadPluginDataJSON();
            OnEnable();
            BuildFilterListFromSceneAtoms();
        }
        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
    }

    private void BuildFilterListFromSceneAtoms()
    {
        filters = new List<Filter>();
        List<Atom> atoms = GetSceneAtoms();
        foreach (Atom atom in atoms)
        {
            // cannot recommend to *not* exclude person atom
            // noticed weird behavior with clothing if hidden
            if (alwaysVisibleAtomTypes.Contains(atom.type))
                continue;
            // do not create a filter for CUAs linked to a Rigidbody (tied to person)
            // they will be excluded by default in ShowAtom()
            if (atom.type.Equals("CustomUnityAsset") && atom.mainController.linkToRB != null)
                continue;
            Filter filter = new Filter();
            filter.atom = atom.type;
            filter.show = new JSONStorableBool(atom.name, false);
            filter.show.setCallbackFunction = (b) => { ShowAtom(atom, b, false); };
            filter.toggle = CreateToggle(filter.show, true);
            filter.toggle.gameObject.SetActive(false);
            filters.Add(filter);
        }
    }

    void Update()
    {
        if (Input.GetKey("left ctrl") || Input.GetKey("right ctrl"))
        {
            if (Input.GetKeyDown("r"))
                jRenderers.val = !jRenderers.val;
        }
    }

    // Required for Clear- and Remove-buttons
    private void OnDestroy() { OnDisable(); }

    private void OnEnable()
    {
        Log("OnEnable()");
        isEnabled = true;
        if (!initOK)
            return;
        SetColor(jColor.val.H, jColor.val.S, jColor.val.V);
        skyUI.showSkyboxToggle.isOn = jShowSkybox.val;
        ShowRenderers(jRenderers.val);
        // Activate UI
        colorPicker.gameObject.SetActive(true);
        buttonSave.gameObject.SetActive(true);
        buttonLoad.gameObject.SetActive(true);
        buttonDelete.gameObject.SetActive(true);
        toggleSkybox.gameObject.SetActive(true);
        toggleRenderers.gameObject.SetActive(true);
        textField.text = TEXT;
    }

    private void OnDisable()
    {
        Log("OnDisable()");
        if (!initOK)
            return;
        SetColor(0f, 0f, 0f);
        skyUI.showSkyboxToggle.isOn = jOriginalSkyboxState.val;
        ShowRenderers(true);
        isEnabled = false; // Disable after color/skybox reset!
        // Deactivate UI
        colorPicker.gameObject.SetActive(false);
        buttonSave.gameObject.SetActive(false);
        buttonLoad.gameObject.SetActive(false);
        buttonDelete.gameObject.SetActive(false);
        toggleSkybox.gameObject.SetActive(false);
        toggleRenderers.gameObject.SetActive(false);
        textField.text = "Plugin disabled.\nEnable to access user interface.";
    }

    private void SetColor(float h, float s, float v)
    {
        Log("SetColor()");
        if (!initOK || !isStarted || !isEnabled)
            return;
        mainCam.backgroundColor = Color.HSVToRGB(h, s, v);
        screenshotCam.backgroundColor = mainCam.backgroundColor;
        thumbnailCam.backgroundColor = mainCam.backgroundColor;
        if (windowCam)
            windowCam.backgroundColor = mainCam.backgroundColor;
    }

    private void ShowSkybox(bool b) { if (!initOK || !isStarted || !isEnabled) return; skyUI.showSkyboxToggle.isOn = b; }

    private void ShowRenderers(bool b)
    {
        if (!initOK || !isStarted || !isEnabled || filters == null)
            return;
        try
        {
            foreach (Filter filter in filters)
            {
                filter.toggle.gameObject.SetActive(!b);
            }

            List<Atom> atoms = GetSceneAtoms();
            foreach (Atom atom in atoms)
            {
                ShowAtom(atom, b);
            }
        }
        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
    }

    // Unused
    private void ShowAtomByID(string uid, bool b)
    {
        Atom atom = GetAtomById(uid);
        if (atom == null)
            return;
        ShowAtom(atom, b);
    }

    private void ShowAtom(Atom atom, bool b, bool useFilter = true)
    {
        if (useFilter)
        {
            if (filters == null)
                return;
            Filter filterForCurrentAtom = null;
            foreach (Filter filter in filters)
            {
                if (atom.type.Equals(filter.atom))
                {
                    filterForCurrentAtom = filter;
                    continue;
                }
            }

            if (filterForCurrentAtom != null && filterForCurrentAtom.show.val)
            {
                if (!b)
                {
                    b = true;
                    Log($"forceShowByFilter: {atom.name} of type {atom.type}");
                    goto forceShowByFilter;
                }
                Log($"ignored {atom.name} of type {atom.type} (user filter)");
                return;
            }
            // Exclude atoms in alwaysVisibleAtomTypes-list
            if (alwaysVisibleAtomTypes.Contains(atom.type))
            {
                Log($"ignored {atom.name} of type {atom.type} (always visible)");
                return;
            }
        }

        // Exclude atoms of type "CustomUnityAsset" with valid link to a Ridigbody
        // This should keep CUA visible that are linked to a person
        if (atom.type.Equals("CustomUnityAsset") && atom.mainController.linkToRB != null)
        {
            Log($"ignored {atom.name} of type {atom.type} (Rigidbody-link)");
            return;
        }

        forceShowByFilter:

        // Hide "Renderer" objects
        Renderer[] renderers = atom.GetComponentsInChildren<Renderer>();
        foreach (Renderer r in renderers)
        {
            if (jDebug.val)
                SuperController.LogMessage($"{r.GetType()} on atom {atom.type} = {b}");
            r.enabled = b;
        }

        // Hide "MirrorReflection", "ReflectiveSlate" uses this & MeshRenderer
        MirrorReflection[] reflections = atom.GetComponentsInChildren<MirrorReflection>();
        foreach (MirrorReflection reflection in reflections)
        {
            if (jDebug.val)
                SuperController.LogMessage($"{reflection.GetType()} on atom {atom.type} = {b}");
            reflection.enabled = b;
        }

        // Hide "DAZSkinV2", used by "TechnoLight"-atom
        DAZSkinV2[] skins = atom.GetComponentsInChildren<DAZSkinV2>();
        foreach (DAZSkinV2 skin in skins)
        {
            if (jDebug.val)
                SuperController.LogMessage($"{skin.GetType()} on atom {atom.type} = {b}");
            skin.enabled = b;
            //skin.draw = b; // alternative
        }

        // Hide "CanvasRenderer", used by "WebPanelEmissive"-atom
        CanvasRenderer[] canvasRenderers = atom.GetComponentsInChildren<CanvasRenderer>();
        foreach (CanvasRenderer canvasRenderer in canvasRenderers)
        {
            if (jDebug.val)
                SuperController.LogMessage($"{canvasRenderer.GetType()} on atom {atom.type} = {b}");
            if (b)
                canvasRenderer.SetAlpha(1f);
            else
                canvasRenderer.SetAlpha(0f);
        }

        // Hide "Terrain", used in some scenes by Meshed
        // In Visual Studio 'Terrain' is undefined for me and throws an error.
        // Appearently because it's not part of the assembly or something.
        // But it will compile by VaM! (Used Meshed VS Plugin Project template)
        UnityEngine.Terrain[] terrains = atom.GetComponentsInChildren<UnityEngine.Terrain>();
        foreach (UnityEngine.Terrain terrain in terrains)
        {
            if (jDebug.val)
                SuperController.LogMessage($"{terrain.GetType()} on atom {atom.type} = {b}");
            terrain.enabled = b;
        }
    }

    private void SavePluginDataJSON()
    {
        FileManagerSecure.CreateDirectory(SAVE_FOLDER);
        JSONClass jc = new JSONClass
            {
                {"hue", jColor.val.H.ToString()},
                {"sat", jColor.val.S.ToString()},
                {"val", jColor.val.V.ToString()},
            };

        string path = SAVE_FOLDER + "\\" + SAVE_FILENAME + "." + SAVE_EXTENSION;
        SaveJSON(jc, path);
        SuperController.LogMessage("CameraBackgroundColor saved " + path);
    }

    private void LoadPluginDataJSON()
    {
        string path = SAVE_FOLDER + "\\" + SAVE_FILENAME + "." + SAVE_EXTENSION;
        if (!FileManagerSecure.FileExists(path))
            // silent fail if user has no default file saved
            return;
        JSONClass jc = LoadJSON(path).AsObject;
        if (jc == null)
        {
            SuperController.LogMessage("CameraBackgroundColor failed to load user default - LoadJSON() == null");
            return;
        }
        if (!jc.HasKey("hue") || !jc.HasKey("sat") || !jc.HasKey("val"))
        {
            SuperController.LogMessage("CameraBackgroundColor failed to load user default - missing key(s), file modified?");
            return;
        }
        jColor.SetVal(jc["hue"].AsFloat, jc["sat"].AsFloat, jc["val"].AsFloat);
    }

    private void DeletePluginDataJSON()
    {
        string path = SAVE_FOLDER + "\\" + SAVE_FILENAME + "." + SAVE_EXTENSION;
        FileManagerSecure.DeleteFile(path);
        SuperController.LogMessage("CameraBackgroundColor deleted " + path);
    }

    private void CheckForMultipleLoadedScripts()
    {
        List<Atom> atoms = GetSceneAtoms();
        List<Atom> atomsWithScript = new List<Atom>();
        foreach (Atom atom in atoms)
        {
            MVRScript[] scripts = atom.GetComponentsInChildren<MVRScript>();
            foreach (MVRScript script in scripts)
                if (script.name.Contains("CameraBackgroundColor"))
                    atomsWithScript.Add(atom);
        }

        if (atomsWithScript.Count > 1)
        {
            // Found multiple scripts -> iterate over atoms with script loaded, print their names
            foreach (Atom atom in atomsWithScript)
                if (atom.name == "CoreControl")
                    SuperController.LogMessage("CameraBackgroundColor on atom: CoreControl  (Session-Plugins- or Scene-Plugins-tab)");
                else
                    SuperController.LogMessage("CameraBackgroundColor on atom: " + atom.name);
            SuperController.LogMessage("WARNING: multiple CameraBackgroundColor-Plugins loaded. Remove others to avoid conflict.");
        }
    }

    private void OnClickSupportButton()
    {
        Application.OpenURL("https://www.patreon.com/sallyw/about");
    }

    public void Log(string s)
    {
        // Was debug toggle created already and is it's default state on?
        if (jDebug == null)
        {
            if (debugToggleDefaultState)
            {
                SuperController.LogMessage(s);
            }
        }
        // print debug message based on toggle state
        else if (jDebug.val) // debug toggle on?
        {
            SuperController.LogMessage(s);
        }
    }
}